Unit SIDFM;

{ 9 voice FM SID !

 FM   Emulate

  0 - SID0 TRI/NSE
  1 - SID0 SAW
  2 - SID0 SQR
  3 - SID1 TRI/NSE
  4 - SID1 SAW
  5 - SID1 SQR
  6 - SID2 TRI/NSE
  7 - SID2 SAW
  8 - SID2 SQR


}

Interface

uses DATA;

Procedure Init;
Procedure Done;

Procedure PokeSID; { called by POKE.ASM (set SIDAddr & SIDPoke }
Procedure ResetFM; { called by CBM64.ASM }

Procedure delay(ms : word);

Implementation

Const
 AdlibFM=$388;

 Op:array[0..8] of word = ( { i mod 3 + (i div 3) shl 3 }
  $000,$001,$002,$008,$009,$00A,$010,$011,$012
 );

 Waves:array[0..8] of byte=(0,3,1,0,3,1,0,3,1); { TRI/SAW/SQR... }

Var
 Percussive:byte;
 MaxVol:byte;
 Voice2:boolean;

 Notes:array[0..8] of byte;
 State:array[0..8] of byte;
 Freqs:array[0..8] of word;
 Pulse:array[0..8] of word;

Procedure delay(ms : word); assembler; { SWAG }
{ms is the number of milliseconds to delay. 1000ms = 1second}
 asm
  mov ax,1000
  mul ms
  jz  @done
  mov cx,dx
  mov dx,ax
  mov ah,$86
  int $15
  @done:
 end;

Procedure WriteFM(reg,val:word); assembler;
 asm
  mov dx,AdlibFM; mov ax,reg; out dx,al
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  inc dx; mov ax,Val; out dx,al
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
 end;

Function ReadFM:byte; assembler;
 asm
  mov dx,AdlibFM
  in al,dx
 end;

Procedure ResetFM;
 var
  r,v,Oper,w:byte;
 begin
  (* zero all registers *)
  for r:=$01 to $F5 do WriteFM(r,0);

  (* allow FM chips to control the waveform of each operator *)
  WriteFM($01,$20);
{  WriteFM($08,$80); { ??? }

  FillChar(Notes,SizeOf(Notes),0);
  FillChar(State,SizeOf(Notes),0);
  FillChar(Freqs,SizeOf(Freqs),0);
  FillChar(Pulse,SizeOf(Pulse),0);

  for v:=0 to 8 do begin
   Oper:=Op[v];
   w:=Waves[v];
   WriteFM($E0+Oper,w);
   WriteFM($E3+Oper,w);
   WriteFM($20+Oper,$21);
   WriteFM($23+Oper,$21);

   WriteFM($C0+v,$01);
  end;
 end;

Procedure Init;
 var
  s1,s2:byte;
  r:byte;
 begin
  WriteFM($04,$60);   (* reset both timers        *)
  WriteFM($04,$80);   (* enable timer interrupts  *)
  s1:=ReadFM;         (* read status register     *)
  WriteFM($02,$FF);
  WriteFM($04,$21);   (* start timer 1            *)
  Delay(80);          (* could do something useful*)
  s2:=ReadFM;         (* read status register     *)
  WriteFM($04,$60);   (* reset both timers        *)
  WriteFM($04,$80);   (* enable timer interrupts  *)

  if (((s1 and $E0)<>0)or((s2 and $E0)<>$C0)) then begin
   WriteLn(#16' no AdlibFM (388h) found');
   halt;
  end;

  ResetFM;

  WriteLn(#16' AdlibFM Found at 388h');
 end;

Procedure Done;
 var
  i:integer;
 begin
  ResetFM;
  for i:=0 to 15 do WriteFM($40+i,63); { no sound }
  for i:=0 to 15 do WriteFM($B0+i,0);  { no freq  }
 end;

Procedure PlayVoice(Voice:word);
 var
  Wave:word;
  Freq:word;
  Puls:word;
  Note:byte;
  Oper:byte;
 begin
  Wave:=Waves[Voice]; { Wave Form, SQR need a Pulse Volume ? }
  Puls:=Pulse[Voice]; { PulseWidth }
  Note:=Notes[Voice]; { Note On or Off ($20 or $00) }

  Oper:=Op[Voice];

  if Note>0 then begin
  {
   Freq:=State[Voice];
   WriteFM($A0+Voice,Lo(Freq));
   WriteFM($B0+Voice,Hi(Freq));
  }
   if Wave=1 then WriteFM($40+Oper,Puls) else WriteFM($40+Oper,MaxVol);
   if Wave=1 then WriteFM($43+Oper,Puls) else WriteFM($43+Oper,MaxVol);
   Freq:=Freqs[Voice];
   State[Voice]:=Freq;
  end else begin
   Freq:=State[Voice];
   State[Voice]:=0;
   if Freq=0 then exit;
  end;

  WriteFM($A0+Voice,Lo(Freq));
  WriteFM($B0+Voice,Hi(Freq) or Note);

 end;

Procedure VoiceFreq(Voice,Freq:word);
 var
  oct:byte;
 begin
 { Adlib Freq from C64 Freq :

   c64   = F*17.0284
   Adlib = F/0.763

   Adlib = c64/13

 }
  Freq:=(Freq div 13);
  Oct:=4;
 { Ajdust Octave }
  while Freq>$2AE do begin { C4 }
   inc(oct);
   freq:=freq shr 1;
  end;
  while (oct>0)and(Freq<$16B) do begin { C#4 }
   dec(oct);
   freq:=freq shl 1;
  end;
  Freq:=Freq or (Oct shl (2+8));

  Voice:=3*Voice;
  Freqs[Voice]  :=Freq;
  Freqs[Voice+1]:=Freq;
  Freqs[Voice+2]:=Freq;

  PlayVoice(Voice);
  PlayVoice(Voice+1);
  PlayVoice(Voice+2);
 end;

Const
 PulseMod:array[0..31] of byte=(
   00,01,02,03,04,04,05,05,
   06,06,07,07,08,08,09,09,
   10,10,11,11,12,12,13,13,
   14,14,15,15,16,16,16,16
 );

Procedure VoicePulse(Voice,APulse:word);
 begin
  APulse:=PulseMod[APulse shr 7]; { ??? from PC64 }
  Voice:=3*Voice+2; { Only SQR }
  Pulse[Voice]:=APulse;
  PlayVoice(Voice);
 end;

Procedure VoiceControl(Voice:word);
 var
  Wave:word;
  Back:word;
  Freq:word;
 begin

  Wave:=SIDPoke shr 4;

  if (Voice=2) and (Not Voice2) then Wave:=9; { Mute 2 }
  if (Wave=0) then Wave:=9; { Invalide ? }

  Voice:=3*Voice;
  Notes[Voice]  :=$00;
  Notes[Voice+1]:=$00;
  Notes[Voice+2]:=$00;

  if (Wave<9) and Odd(SIDPoke) then begin

   if Wave=8 then begin { Noise }

    WriteFM($C0+Voice,$0E);
    Notes[Voice]:=$20;

   end else begin

    WriteFM($C0+Voice,$01);

    { Wave:=1+2+4; { Try this ! }

    if (Wave and 1)>0 then Notes[Voice]  :=$20;
    if (Wave and 2)>0 then Notes[Voice+1]:=$20;
    if (Wave and 4)>0 then Notes[Voice+2]:=$20;

   end;

  end;

  PlayVoice(Voice);
  PlayVoice(Voice+1);
  PlayVoice(Voice+2);
 end;

Const
 Attack:array[0..15] of byte=(
  16*(15-4 ),16*(15-6 ),16*(15-7 ),16*(15-8),
  16*(15-8 ),16*(15-9 ),16*(15-9 ),16*(15-9),
  16*(15-10),16*(15-11),16*(15-12),16*(15-13),
  16*(15-13),16*(15-14),16*(15-14),16*(15-14)
 );
 Decay:array[0..15] of byte=(
  15-4 ,15-5 ,15-6 ,15-6,
  15-6 ,15-6 ,15-6 ,15-7,
  15-8 ,15-9 ,15-10,15-11,
  15-12,15-13,15-14,15-14
 );
 Sustain:array[0..15] of byte=(
  16*(15-0 ),16*(15-8 ),16*(15-10),16*(15-11),
  16*(15-12),16*(15-12),16*(15-13),16*(15-13),
  16*(15-13),16*(15-14),16*(15-14),16*(15-14),
  16*(15-14),16*(15-14),16*(15-15),16*(15-15)
 );
 Release:array[0..15] of byte=(
  15-0 ,15-1 ,15-2 ,15-3,
  15-4 ,15-5 ,15-6 ,15-7,
  15-8 ,15-9 ,15-10,15-11,
  15-12,15-13,15-14,15-14
 );

Procedure VoiceAD(Voice:word);
 var
  ATDC:byte;
  V,O:word;
 begin
  ATDC:=Attack[SIDPoke shr 4] or Decay[SIDPoke and 15];

  Voice:=3*Voice;
  For V:=Voice to Voice+2 do begin
   o:=op[V];
   WriteFM($60+O,ATDC);
   WriteFM($63+O,ATDC);
  end;
 end;

Procedure VoiceSR(Voice:word);
 var
  STRL:byte;
  V,O:word;
 begin
  STRL:=Sustain[SIDPoke shr 4] or Release[SIDPoke and 15];
  Voice:=3*Voice;
  for V:=Voice to Voice+2 do begin
   O:=Op[V];
   WriteFM($80+O,STRL);
   WriteFM($83+O,STRL);
  end;
 end;

Procedure SetVolume;
 var
  v:byte;
 begin
  Voice2:=(SIDPoke and $80)=0;
  MaxVol:=(SIDPoke and 15);
  MaxVol:=63-(2*MaxVol+32);
  for v:=0 to 8 do PlayVoice(v);
 end;


Procedure PokeSID;
 begin
  case SIDAddr of
{ VOICE 0 }
   0,1: VoiceFreq(0,MemW[CIOSeg:$D400+0]);
   2,3: VoicePulse(0,MemW[CIOSeg:$D400+2] and $FFF);

   4: VoiceControl(0);
   5: VoiceAD(0);
   6: VoiceSR(0);

{ VOICE 1 }
   7,8 : VoiceFreq(1,MemW[CIOSeg:$D400+7]);
   9,10: VoicePulse(1,MemW[CIOSeg:$D400+9] and $FFF);

   11: VoiceControl(1);
   12: VoiceAD(1);
   13: VoiceSR(1);

{ VOICE 2 }
   14,15: VoiceFreq(2,MemW[CIOSeg:$D400+14]);
   16,17: VoicePulse(2,MemW[CIOSeg:$D400+16] and $FFF);

   18: VoiceControl(2);
   19: VoiceAD(2);
   20: VoiceSR(2);

{ GENERAL }
{   21,22: FilterFreq:=MemW[CIOSeg:$D400+21]; }
{   23   : SetFilter; }
    24   : SetVolume;
  end;
 end;
end.